# -*- coding: utf-8 -*-
# BELA's package version information

# This code is a part of BELA package: https://github.com/letuananh/bela
# :developer: Le Tuan Anh <tuananh.ke@gmail.com>
# :license: MIT, see LICENSE for more details.

__author__ = "Le Tuan Anh"
__email__ = "tuananh.ke@gmail.com"
__copyright__ = "Copyright (c) 2019, Le Tuan Anh, BLIP lab NTU, Singapore"
__credits__ = []
__license__ = "MIT License"
__description__ = "BLIP ELAN Language Annotation package"
__url__ = "https://bela.readthedocs.io/"
__maintainer__ = "Le Tuan Anh"
__version_major__ = "2.0.0"
__version__ = "{}a21.post6".format(__version_major__)
__version_long__ = "{} - Alpha 21.post6".format(__version_major__)
__status__ = "3 - Alpha"
